import datetime
from dataclasses import dataclass
from typing import Self

import httpx
from dateutil.parser import parse


@dataclass
class SimpleResponseHeaders:
    content_type: str
    etag: str | None
    last_modified: datetime.datetime | None
    last_checked: datetime.datetime | None

    @classmethod
    def from_raw_headers(cls, data: httpx.Headers) -> Self | None:
        if (pagetype := data.get("content-type")) and "text/html" in pagetype:
            checked = data.get("date")
            modified = data.get("last-modified")
            return cls(
                content_type=data["content-type"],
                etag=data.get("etag", None),
                last_modified=parse(modified) if modified else None,
                last_checked=parse(checked) if checked else None,
            )
        return None



from dataclasses import asdict, dataclass
from http import HTTPStatus
from urllib.parse import urlparse

import httpx
from bs4 import BeautifulSoup

from .headers import SimpleResponseHeaders
from .meta import SimpleMeta
from .repo import GithubRepo


@dataclass
class InspectedURL:
    url: str
    head: httpx.Headers | None = None
    content: bytes | None = None

    def __post_init__(self):
        r = httpx.get(self.url, follow_redirects=True)
        if not r.status_code == HTTPStatus.OK:
            raise Exception(f"Bad {self.url}: {r.status_code}")
        self.head = r.headers
        self.content = r.content

    def prep(self, dc) -> dict:
        """Must be a dataclass."""
        return {dc.__class__.__name__: asdict(dc)} if dc is not None else {}

    @property
    def parsed_url(self) -> dict:
        """Generates the parts of urlparse.ParsedResult as a dictionary."""
        parsed = urlparse(self.url)
        return {
            "ParsedResult": {k: getattr(parsed, k) for k in parsed._fields}
        }

    @property
    def site_headers(self) -> SimpleResponseHeaders | None:
        """Selected HTML headers, requires content-type to include `text/html`."""
        if self.head:
            return SimpleResponseHeaders.from_raw_headers(data=self.head)

    @property
    def site_meta(self) -> SimpleMeta | None:
        """Simplified summary of meta tags."""
        if self.content:
            html_soup = BeautifulSoup(self.content, "html.parser")
            return SimpleMeta.from_soup(html_soup)

    @property
    def repo_meta(self) -> GithubRepo | None:
        """Usable fields from Github repository, requires Github token."""
        if raw_data := GithubRepo.matches(self.url):
            return GithubRepo(**raw_data)

    @property
    def export(self):
        """Combine url, header, meta, repo key-value pairs."""
        data = self.parsed_url
        head = self.prep(self.site_headers) or {}
        meta = self.prep(self.site_meta) or {}
        repo = self.prep(self.repo_meta) or {}
        return data | head | meta | repo



import re
from dataclasses import dataclass, field, fields
from http import HTTPStatus
from urllib.parse import urlparse

from start_sdk import Github

gh = Github()


REPO_PATTERN = re.compile(r"^\/(?P<author>.*)\/(?P<repo>.*)$")
REPO_FIELDS = [
    "created_at",
    "updated_at",
    "pushed_at",
    "stargazers_count",
    "description",
]


@dataclass
class Topic:
    name: str


@dataclass
class Owner:
    login: str
    avatar_url: str
    gravatar_id: str

    @classmethod
    def from_github_data(cls, data: dict):
        if (own := data.get("owner")) and isinstance(own, dict):
            keys = [k.name for k in fields(cls)]
            return cls(**{k: v for k, v in own.items() if k in keys})
        return None


@dataclass
class GithubRepo:
    """Requires a private access token (PAT) from Github Developer Settings."""

    author: str
    repo: str
    key: str | None = None
    owner: Owner | None = None
    topics: list[Topic] = field(default_factory=list)
    columns: dict = field(default_factory=dict)

    def __post_init__(self):
        data = self.get_data(author=self.author, repo=self.repo)
        self.key = data.get("license") and data["license"].get("key")
        self.owner = Owner.from_github_data(data)
        self.topics = [Topic(name=name) for name in data.get("topics")]
        self.columns = {k: v for k, v in data.items() if k in REPO_FIELDS}

    @classmethod
    def get_data(cls, author: str, repo: str):
        res = gh.get_repo(author=author, repo=repo)
        if res.status_code != HTTPStatus.OK:
            raise Exception(f"Fail github: {author=} {repo=}")
        return res.json()

    @classmethod
    def matches(cls, url: str) -> dict | None:
        p = urlparse(url)
        if p.netloc == "github.com" and p.path and p.path != "/":
            if match := REPO_PATTERN.search(p.path):
                return match.groupdict()
        return None

    @classmethod
    def from_url(cls, url: str):
        return cls(**d) if (d := cls.matches(url)) else None



from dataclasses import dataclass
from typing import Iterable

from bs4 import BeautifulSoup

TITLE = (
    'meta[name="twitter:title"]',
    'meta[property="og:title"]',
    "title",
)
DESC = (
    'meta[name="twitter:description"]',
    'meta[property="og:description"]',
    'meta[name="description"]',
)
IMG = (
    'meta[name="twitter:image"]',
    'meta[property="og:image"]',
)
AUTHOR = (
    'meta[name="author"]',
    'meta[name="twitter:creator"]',
)
TYPE = ('meta[property="og:type"]',)


@dataclass
class SimpleMeta:
    """Extract generic website metadata based on a url fetched on a certain date.

    All of the fields, except the date, default to `None`.

    Field | Type | Description
    :--:|:--:|:--
    title | str | First matching title parsed from `<meta>` CSS selectors (and the `<title>` tag)
    description | str | First matching description Parsed from `<meta>` CSS selectors
    author | str | Either the author or the creator, if the author is absent
    image | str | An [open graph](https://ogp.me/) (OG) image url detected
    category | str | A type detected from OG ("og:type") values
    """  # noqa: E501

    title: str | None = None
    description: str | None = None
    author: str | None = None
    image: str | None = None
    category: str | None = None

    @classmethod
    def from_soup(cls, soup: BeautifulSoup):
        return cls(
            title=cls.select(soup, TITLE),
            description=cls.select(soup, DESC),
            author=cls.select(soup, AUTHOR),
            image=cls.select(soup, IMG),
            category=cls.select(soup, TYPE),
        )

    @classmethod
    def select(
        cls, soup: BeautifulSoup, selectors: Iterable[str]
    ) -> str | None:
        """The order of CSS selectors. The first one
        matched, retrieves the content, if found.

        See present list of selectors used to extract content:

        ```py
        TITLE = (
            'meta[name="twitter:title"]',
            'meta[property="og:title"]',
            "title",
        )
        DESC = (
            'meta[name="twitter:description"]',
            'meta[property="og:description"]',
            'meta[name="description"]',
        )
        IMG = (
            'meta[name="twitter:image"]',
            'meta[property="og:image"]',
        )
        AUTHOR = (
            'meta[name="author"]',
            'meta[name="twitter:creator"]',
        )
        TYPE = ('meta[property="og:type"]',)
        ```

        Note the special rule on `title` as a selector.

        Examples:
            >>> from pathlib import Path
            >>> html = Path(__file__).parent.parent / "tests" / "data" / "test.html"
            >>> soup = BeautifulSoup(html.read_text(), "html.parser")
            >>> SimpleMeta.select(soup, TITLE)
            'Hello World From Twitter Title!'
            >>> SimpleMeta.select(soup, DESC)
            'this is a description from twitter:desc'

        Args:
            soup (BeautifulSoup): Converted html content into a soup object
            selectors (Iterable[str]): CSS selectors as a tuple

        Returns:
            str | None: If found, return the text value.
        """
        for selector in selectors:
            if selector.startswith("meta"):
                if desc := soup.select(selector):
                    if content := desc[0].get("content"):
                        if content and isinstance(content, str):
                            return content
            elif selector == "title":
                if titles := soup("title"):
                    return titles[0].get_text()
        return None
